@@ -1,7 +1,7 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- 
- *  gtksourceiter.h
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*- */
+/*  gtksourceiter.c
  *
- *  Copyright (C) 2000 - 2005 Paolo Maggi 
+ *  Copyright (C) 2000 - 2005 Paolo Maggi
  *  Copyright (C) 2002, 2003 Jeroen Zwartepoorte
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -30,6 +30,13 @@
 #include <string.h>
 #include "gtksourceiter.h"
 
+/**
+ * SECTION:iter
+ * @Short_description: Improved search support in a #GtkSourceBuffer
+ * @Title: GtkSourceIter
+ * @See_also: #GtkTextIter
+ */
+
 #define GTK_TEXT_UNKNOWN_CHAR 0xFFFC
 
 /* this function acts like g_utf8_offset_to_pointer() except that if it finds a
@@ -56,8 +63,30 @@ pointer_from_offset_skipping_decomp (const gchar *str, gint offset)
 	return p;
 }
 
+static gboolean
+exact_prefix_cmp (const gchar *string,
+		  const gchar *prefix,
+		  guint        prefix_len)
+{
+	GUnicodeType type;
+
+	if (strncmp (string, prefix, prefix_len) != 0)
+		return FALSE;
+	if (string[prefix_len] == '\0')
+		return TRUE;
+
+	type = g_unichar_type (g_utf8_get_char (string + prefix_len));
+
+	/* If string contains prefix, check that prefix is not followed
+	 * by a unicode mark symbol, e.g. that trailing 'a' in prefix
+	 * is not part of two-char a-with-hat symbol in string. */
+	return type != G_UNICODE_COMBINING_MARK &&
+		type != G_UNICODE_ENCLOSING_MARK &&
+		type != G_UNICODE_NON_SPACING_MARK;
+}
+
 static const gchar *
-g_utf8_strcasestr (const gchar *haystack, const gchar *needle)
+utf8_strcasestr (const gchar *haystack, const gchar *needle)
 {
 	gsize needle_len;
 	gsize haystack_len;
@@ -95,7 +124,7 @@ g_utf8_strcasestr (const gchar *haystack, const gchar *needle)
 
 	while (*p)
 	{
-		if ((strncmp (p, needle, needle_len) == 0))
+		if (exact_prefix_cmp (p, needle, needle_len))
 		{
 			ret = pointer_from_offset_skipping_decomp (haystack, i);
 			goto finally_1;
@@ -112,7 +141,7 @@ finally_1:
 }
 
 static const gchar *
-g_utf8_strrcasestr (const gchar *haystack, const gchar *needle)
+utf8_strrcasestr (const gchar *haystack, const gchar *needle)
 {
 	gsize needle_len;
 	gsize haystack_len;
@@ -150,7 +179,7 @@ g_utf8_strrcasestr (const gchar *haystack, const gchar *needle)
 
 	while (p >= caseless_haystack)
 	{
-		if (strncmp (p, needle, needle_len) == 0)
+		if (exact_prefix_cmp (p, needle, needle_len))
 		{
 			ret = pointer_from_offset_skipping_decomp (haystack, i);
 			goto finally_1;
@@ -167,8 +196,8 @@ finally_1:
 }
 
 static gboolean
-g_utf8_caselessnmatch (const char *s1, const char *s2,
-		       gssize n1, gssize n2)
+utf8_caselessnmatch (const char *s1, const char *s2,
+		     gssize n1, gssize n2)
 {
 	gchar *casefold;
 	gchar *normalized_s1;
@@ -200,11 +229,30 @@ g_utf8_caselessnmatch (const char *s1, const char *s2,
 
 finally_2:
 	g_free (normalized_s1);
-	g_free (normalized_s2);	
+	g_free (normalized_s2);
 
 	return ret;
 }
 
+/* FIXME: total horror */
+static gboolean
+char_is_invisible (const GtkTextIter *iter)
+{
+	GSList *tags;
+	gboolean invisible = FALSE;
+	tags = gtk_text_iter_get_tags (iter);
+	while (tags)
+	{
+		gboolean this_invisible, invisible_set;
+		g_object_get (tags->data, "invisible", &this_invisible,
+			      "invisible-set", &invisible_set, NULL);
+		if (invisible_set)
+			invisible = this_invisible;
+		tags = g_slist_delete_link (tags, tags);
+	}
+	return invisible;
+}
+
 static void
 forward_chars_with_skipping (GtkTextIter *iter,
 			     gint         count,
@@ -231,24 +279,28 @@ forward_chars_with_skipping (GtkTextIter *iter,
 		if (skip_nontext && gtk_text_iter_get_char (iter) == GTK_TEXT_UNKNOWN_CHAR)
 			ignored = TRUE;
 
-		if (!ignored && skip_invisible &&
-		    /* _gtk_text_btree_char_is_invisible (iter)*/ FALSE)
+		/* FIXME: char_is_invisible() gets list of tags for each char there,
+		   and checks every tag. It doesn't sound like a good idea. */
+		if (!ignored && skip_invisible && char_is_invisible (iter))
 			ignored = TRUE;
 
 		if (!ignored && skip_decomp)
 		{
 			/* being UTF8 correct sucks; this accounts for extra
 			   offsets coming from canonical decompositions of
-			   UTF8 characters (e.g. accented characters) which 
+			   UTF8 characters (e.g. accented characters) which
 			   g_utf8_normalize() performs */
 			gchar *normal;
+			gchar *casefold;
 			gchar buffer[6];
 			gint buffer_len;
 
 			buffer_len = g_unichar_to_utf8 (gtk_text_iter_get_char (iter), buffer);
-			normal = g_utf8_normalize (buffer, buffer_len, G_NORMALIZE_NFD);
+			casefold = g_utf8_casefold (buffer, buffer_len);
+			normal = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
 			i -= (g_utf8_strlen (normal, -1) - 1);
 			g_free (normal);
+			g_free (casefold);
 		}
 
 		gtk_text_iter_forward_char (iter);
@@ -304,15 +356,15 @@ lines_match (const GtkTextIter *start,
 
 	if (match_start) /* if this is the first line we're matching */
 	{
-		found = g_utf8_strcasestr (line_text, *lines);
+		found = utf8_strcasestr (line_text, *lines);
 	}
 	else
 	{
 		/* If it's not the first line, we have to match from the
 		 * start of the line.
 		 */
-		if (g_utf8_caselessnmatch (line_text, *lines, strlen (line_text),
-					   strlen (*lines)))
+		if (utf8_caselessnmatch (line_text, *lines, strlen (line_text),
+					 strlen (*lines)))
 			found = line_text;
 		else
 			found = NULL;
@@ -402,15 +454,15 @@ backward_lines_match (const GtkTextIter *start,
 
 	if (match_start) /* if this is the first line we're matching */
 	{
-		found = g_utf8_strrcasestr (line_text, *lines);
+		found = utf8_strrcasestr (line_text, *lines);
 	}
 	else
 	{
 		/* If it's not the first line, we have to match from the
 		 * start of the line.
 		 */
-		if (g_utf8_caselessnmatch (line_text, *lines, strlen (line_text),
-					   strlen (*lines)))
+		if (utf8_caselessnmatch (line_text, *lines, strlen (line_text),
+					 strlen (*lines)))
 			found = line_text;
 		else
 			found = NULL;
@@ -454,9 +506,9 @@ backward_lines_match (const GtkTextIter *start,
 
 /* strsplit () that retains the delimiter as part of the string. */
 static gchar **
-strbreakup (const char *string,
-	    const char *delimiter,
-	    gint        max_tokens)
+breakup_string (const char *string,
+		const char *delimiter,
+		gint        max_tokens)
 {
 	GSList *string_list = NULL, *slist;
 	gchar **str_array, *s, *casefold, *new_string;
@@ -519,17 +571,17 @@ strbreakup (const char *string,
  * @iter: start of search.
  * @str: a search string.
  * @flags: flags affecting how the search is done.
- * @match_start: return location for start of match, or %%NULL.
- * @match_end: return location for end of match, or %%NULL.
- * @limit: bound for the search, or %%NULL for the end of the buffer.
- * 
- * Searches forward for @str. Any match is returned by setting 
- * @match_start to the first character of the match and @match_end to the 
+ * @match_start: return location for start of match, or %NULL.
+ * @match_end: return location for end of match, or %NULL.
+ * @limit: bound for the search, or %NULL for the end of the buffer.
+ *
+ * Searches forward for @str. Any match is returned by setting
+ * @match_start to the first character of the match and @match_end to the
  * first character after the match. The search will not continue past
  * @limit. Note that a search is a linear or O(n) operation, so you
  * may wish to use @limit to avoid locking up your UI on large
  * buffers.
- * 
+ *
  * If the #GTK_SOURCE_SEARCH_VISIBLE_ONLY flag is present, the match may
  * have invisible text interspersed in @str. i.e. @str will be a
  * possibly-noncontiguous subsequence of the matched range. similarly,
@@ -542,7 +594,7 @@ strbreakup (const char *string,
  *
  * Same as gtk_text_iter_forward_search(), but supports case insensitive
  * searching.
- * 
+ *
  * Return value: whether a match was found.
  **/
 gboolean
@@ -566,7 +618,7 @@ gtk_source_iter_forward_search (const GtkTextIter   *iter,
 	if ((flags & GTK_SOURCE_SEARCH_CASE_INSENSITIVE) == 0)
 		return gtk_text_iter_forward_search (iter, str, flags,
 						     match_start, match_end,
-						     limit); 
+						     limit);
 
 	if (limit && gtk_text_iter_compare (iter, limit) >= 0)
 		return FALSE;
@@ -597,7 +649,7 @@ gtk_source_iter_forward_search (const GtkTextIter   *iter,
 	slice = (flags & GTK_SOURCE_SEARCH_TEXT_ONLY) == 0;
 
 	/* locate all lines */
-	lines = strbreakup (str, "\n", -1);
+	lines = breakup_string (str, "\n", -1);
 
 	search = *iter;
 
@@ -615,8 +667,8 @@ gtk_source_iter_forward_search (const GtkTextIter   *iter,
 		if (lines_match (&search, (const gchar**)lines,
 				 visible_only, slice, &match, &end))
 		{
-			if (limit == NULL || (limit &&
-					      gtk_text_iter_compare (&end, limit) < 0))
+			if (limit == NULL ||
+			    (limit && gtk_text_iter_compare (&end, limit) <= 0))
 			{
 				retval = TRUE;
 
@@ -639,13 +691,13 @@ gtk_source_iter_forward_search (const GtkTextIter   *iter,
  * @iter: a #GtkTextIter where the search begins.
  * @str: search string.
  * @flags: bitmask of flags affecting the search.
- * @match_start: return location for start of match, or %%NULL.
- * @match_end: return location for end of match, or %%NULL.
- * @limit: location of last possible @match_start, or %%NULL for start of buffer.
- * 
+ * @match_start: return location for start of match, or %NULL.
+ * @match_end: return location for end of match, or %NULL.
+ * @limit: location of last possible @match_start, or %NULL for start of buffer.
+ *
  * Same as gtk_text_iter_backward_search(), but supports case insensitive
  * searching.
- * 
+ *
  * Return value: whether a match was found.
  **/
 gboolean
@@ -669,7 +721,7 @@ gtk_source_iter_backward_search (const GtkTextIter   *iter,
 	if ((flags & GTK_SOURCE_SEARCH_CASE_INSENSITIVE) == 0)
 		return gtk_text_iter_backward_search (iter, str, flags,
 						      match_start, match_end,
-						      limit); 
+						      limit);
 
 	if (limit && gtk_text_iter_compare (iter, limit) <= 0)
 		return FALSE;
@@ -700,7 +752,7 @@ gtk_source_iter_backward_search (const GtkTextIter   *iter,
 	slice = (flags & GTK_SOURCE_SEARCH_TEXT_ONLY) == 0;
 
 	/* locate all lines */
-	lines = strbreakup (str, "\n", -1);
+	lines = breakup_string (str, "\n", -1);
 
 	search = *iter;
 
