@@ -30,19 +30,16 @@
 #include <string.h>
 #include <stdlib.h>
 #include <gtk/gtk.h>
+#include <gio/gio.h>
 #include <gtksourceview/gtksourceview.h>
 #include <gtksourceview/gtksourcelanguage.h>
 #include <gtksourceview/gtksourcelanguagemanager.h>
 #include <gtksourceview/gtksourcestyleschememanager.h>
 #include <gtksourceview/gtksourceprintcompositor.h>
+#include <gtksourceview/gtksourceiter.h>
 #ifdef TEST_XML_MEM
 #include <libxml/xmlreader.h>
 #endif
-#ifdef USE_GNOME_VFS
-#include <libgnomevfs/gnome-vfs-init.h>
-#include <libgnomevfs/gnome-vfs-mime-utils.h>
-#include <libgnomevfs/gnome-vfs-utils.h>
-#endif
 
 /* Global list of open windows */
 
@@ -62,8 +59,10 @@ static GtkSourceStyleScheme *style_scheme = NULL;
 static void       open_file_cb                   (GtkAction       *action,
 						  gpointer         user_data);
 static void       print_file_cb                  (GtkAction       *action,
-						  gpointer         user_data);						  
-static void       debug_thing_3_cb		 (GtkAction       *action,
+						  gpointer         user_data);
+static void       find_cb			 (GtkAction       *action,
+						  gpointer         user_data);
+static void       replace_cb			 (GtkAction       *action,
 						  gpointer         user_data);
 
 static void       new_view_cb                    (GtkAction       *action,
@@ -74,8 +73,12 @@ static void       marks_toggled_cb               (GtkAction       *action,
 						  gpointer         user_data);
 static void       margin_toggled_cb              (GtkAction       *action,
 						  gpointer         user_data);
+static void       hl_bracket_toggled_cb          (GtkAction       *action,
+						  gpointer         user_data);
 static void       hl_line_toggled_cb             (GtkAction       *action,
 						  gpointer         user_data);
+static void       draw_spaces_toggled_cb	 (GtkAction       *action,
+						  gpointer         user_data);
 static void       wrap_lines_toggled_cb          (GtkAction       *action,
 						  gpointer         user_data);
 static void       auto_indent_toggled_cb         (GtkAction       *action,
@@ -89,6 +92,13 @@ static void       indent_toggled_cb              (GtkAction       *action,
 						  GtkAction       *current,
 						  gpointer         user_data);
 
+static void       forward_string_cb              (GtkAction       *action,
+						  gpointer         user_data);
+
+static void       backward_string_cb             (GtkAction       *action,
+						  gpointer         user_data);
+
+
 static GtkWidget *create_view_window             (GtkSourceBuffer *buffer,
 						  GtkSourceView   *from);
 
@@ -103,20 +113,29 @@ static GtkActionEntry buffer_action_entries[] = {
 };
 
 static GtkActionEntry view_action_entries[] = {
-	{ "FileMenu", NULL, "_File" },
+	{ "FileMenu", NULL, "_File", NULL, NULL, NULL },
 	{ "Print", GTK_STOCK_PRINT, "_Print", "<control>P",
 	  "Print the current file", G_CALLBACK (print_file_cb) },
-	{ "ViewMenu", NULL, "_View" },
+	{ "ViewMenu", NULL, "_View", NULL, NULL, NULL },
 	{ "NewView", GTK_STOCK_NEW, "_New View", NULL,
 	  "Create a new view of the file", G_CALLBACK (new_view_cb) },
-	{ "TabWidth", NULL, "_Tab Width" },
-	{ "IndentWidth", NULL, "I_ndent Width" },
-	{ "SmartHomeEnd", NULL, "_Smart Home/End" },
-	{ "DebugThing3", GTK_STOCK_FIND_AND_REPLACE, "Search and _Replace", "<control>R",
-	  "Search and Replace", G_CALLBACK (debug_thing_3_cb) },
+	{ "TabWidth", NULL, "_Tab Width", NULL, NULL, NULL },
+	{ "IndentWidth", NULL, "I_ndent Width", NULL, NULL, NULL },
+	{ "SmartHomeEnd", NULL, "_Smart Home/End", NULL, NULL, NULL },
+	{ "Find", GTK_STOCK_FIND, "_Find", "<control>F",
+	  "Find", G_CALLBACK (find_cb) },
+	{ "Replace", GTK_STOCK_FIND_AND_REPLACE, "Search and _Replace", "<control>R",
+	  "Search and Replace", G_CALLBACK (replace_cb) },
+	{ "ForwardString", NULL, "_Forward to string toggle", "<control>S",
+	  "Forward to the start or end of the next string", G_CALLBACK (forward_string_cb) },
+	{ "BackwardString", NULL, "_Backward to string toggle", "<control><shift>S",
+	  "Backward to the start or end of the next string", G_CALLBACK (backward_string_cb) }
 };
 
 static GtkToggleActionEntry toggle_entries[] = {
+	{ "HlBracket", NULL, "Highlight Matching _Bracket", NULL,
+	  "Toggle highlighting of matching bracket",
+	  G_CALLBACK (hl_bracket_toggled_cb), FALSE },
 	{ "ShowNumbers", NULL, "Show _Line Numbers", NULL,
 	  "Toggle visibility of line numbers in the left margin",
 	  G_CALLBACK (numbers_toggled_cb), FALSE },
@@ -129,6 +148,9 @@ static GtkToggleActionEntry toggle_entries[] = {
 	{ "HlLine", NULL, "_Highlight Current Line", NULL,
 	  "Toggle highlighting of current line",
 	  G_CALLBACK (hl_line_toggled_cb), FALSE },
+	{ "DrawSpaces", NULL, "_Draw Spaces", NULL,
+	  "Draw Spaces",
+	  G_CALLBACK (draw_spaces_toggled_cb), FALSE },
 	{ "WrapLines", NULL, "_Wrap Lines", NULL,
 	  "Toggle line wrapping",
 	  G_CALLBACK (wrap_lines_toggled_cb), FALSE },
@@ -179,10 +201,12 @@ static const gchar *view_ui_description =
 "    <menu action=\"ViewMenu\">"
 "      <menuitem action=\"NewView\"/>"
 "      <separator/>"
+"      <menuitem action=\"HlBracket\"/>"
 "      <menuitem action=\"ShowNumbers\"/>"
 "      <menuitem action=\"ShowMarks\"/>"
 "      <menuitem action=\"ShowMargin\"/>"
 "      <menuitem action=\"HlLine\"/>"
+"      <menuitem action=\"DrawSpaces\"/>"
 "      <menuitem action=\"WrapLines\"/>"
 "      <separator/>"
 "      <menuitem action=\"AutoIndent\"/>"
@@ -210,6 +234,9 @@ static const gchar *view_ui_description =
 "        <menuitem action=\"SmartHomeEndAfter\"/>"
 "        <menuitem action=\"SmartHomeEndAlways\"/>"
 "      </menu>"
+"      <separator/>"
+"      <menuitem action=\"ForwardString\"/>"
+"      <menuitem action=\"BackwardString\"/>"
 "    </menu>"
 "  </menubar>"
 "</ui>";
@@ -220,7 +247,8 @@ static const gchar *buffer_ui_description =
 "    <menu action=\"FileMenu\">"
 "      <menuitem action=\"Open\"/>"
 "      <menuitem action=\"Print\"/>"
-"      <menuitem action=\"DebugThing3\"/>"
+"      <menuitem action=\"Find\"/>"
+"      <menuitem action=\"Replace\"/>"
 "      <separator/>"
 "      <menuitem action=\"Quit\"/>"
 "    </menu>"
@@ -247,7 +275,7 @@ error_dialog (GtkWindow *parent, const gchar *msg, ...)
 					 GTK_DIALOG_DESTROY_WITH_PARENT,
 					 GTK_MESSAGE_ERROR,
 					 GTK_BUTTONS_OK,
-					 tmp);
+					 "%s", tmp);
 	g_free (tmp);
 
 	gtk_dialog_run (GTK_DIALOG (dialog));
@@ -305,129 +333,45 @@ remove_all_marks (GtkSourceBuffer *buffer)
 	gtk_source_buffer_remove_source_marks (buffer, &s, &e, NULL);
 }
 
-/* Note this is wrong for several reasons, e.g. g_pattern_match is broken
- * for glob matching. */
 static GtkSourceLanguage *
-get_language_for_filename (const gchar *filename)
+get_language_for_file (GtkTextBuffer *buffer, const gchar *filename)
 {
-	const gchar * const *languages;
-	gchar *filename_utf8;
 	GtkSourceLanguageManager *manager;
+	GtkSourceLanguage *language;
+	GtkTextIter start, end;
+	gchar *text;
+	gchar *content_type;
+	gboolean result_uncertain;
 
-	filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
-	g_return_val_if_fail (filename_utf8 != NULL, NULL);
-
-	manager = gtk_source_language_manager_get_default ();
-	languages = gtk_source_language_manager_get_language_ids (manager);
-
-	while (*languages != NULL)
-	{
-		GtkSourceLanguage *lang;
-		gchar **globs, **p;
-
-		lang = gtk_source_language_manager_get_language (manager,
-								 *languages);
-		g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (lang), NULL);
-		++languages;
-
-		globs = gtk_source_language_get_globs (lang);
-		if (globs == NULL)
-			continue;
-
-		for (p = globs; *p != NULL; p++)
-		{
-			if (g_pattern_match_simple (*p, filename_utf8))
-			{
-				g_strfreev (globs);
-				g_free (filename_utf8);
-
-				return lang;
-			}
-		}
-
-		g_strfreev (globs);
-	}
-
-	g_free (filename_utf8);
-	return NULL;
-}
-
-/* Note this is wrong, because it ignores mime parent types and subtypes.
- * It's fine to use in a simplish program like this, but is unacceptable
- * in a serious text editor. */
-static GtkSourceLanguage *
-get_language_for_mime_type (const gchar *mime)
-{
-	const gchar * const *languages;
-	GtkSourceLanguageManager *manager;
-
-	manager = gtk_source_language_manager_get_default ();
-	languages = gtk_source_language_manager_get_language_ids (manager);
-
-	while (*languages != NULL)
-	{
-		GtkSourceLanguage *lang;
-		gchar **mimetypes, **p;
-
-		lang = gtk_source_language_manager_get_language (manager,
-								 *languages);
-		g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (lang), NULL);
-		++languages;
-
-		mimetypes = gtk_source_language_get_mime_types (lang);
-
-		if (mimetypes == NULL)
-			continue;
-
-		for (p = mimetypes; *p != NULL; p++)
-		{
-			if (strcmp (*p, mime) == 0)
-			{
-				g_strfreev (mimetypes);
-				return lang;
-			}
-		}
-
-		g_strfreev (mimetypes);
-	}
-
-	return NULL;
-}
-
-static GtkSourceLanguage *
-get_language_for_file (const gchar *filename)
-{
-	GtkSourceLanguage *language = NULL;
-
-#if defined(USE_GNOME_VFS)
-	gchar *mime_type;
-	gchar *uri;
-
-	/* I hate this! */
-	if (g_path_is_absolute (filename))
-	{
-		uri = gnome_vfs_get_uri_from_local_path (filename);
-	}
+	gtk_text_buffer_get_start_iter (buffer, &start);
+	if (gtk_text_buffer_get_char_count (buffer) < 1024)
+		gtk_text_buffer_get_end_iter (buffer, &end);
 	else
+		gtk_text_buffer_get_iter_at_offset (buffer, &end, 1024);
+	text = gtk_text_buffer_get_slice (buffer, &start, &end, TRUE);
+
+	content_type = g_content_type_guess (filename,
+					     (guchar*) text,
+					     strlen (text),
+					     &result_uncertain);
+	if (result_uncertain)
 	{
-		gchar *curdir, *path;
-		curdir = g_get_current_dir ();
-		path = g_strconcat (curdir, "/", filename, NULL);
-		g_free (curdir);
-		uri = gnome_vfs_get_uri_from_local_path (path);
-		g_free (path);
+		g_free (content_type);
+		content_type = NULL;
 	}
 
-	if ((mime_type = gnome_vfs_get_mime_type (uri)))
-		language = get_language_for_mime_type (mime_type);
+	manager = gtk_source_language_manager_get_default ();
+	language = gtk_source_language_manager_guess_language (manager,
+							       filename,
+							       content_type);
 
-	g_free (mime_type);
-	g_free (uri);
-#endif
-
-	if (!language)
-		language = get_language_for_filename (filename);
+	g_message ("Detected '%s' mime type for file %s, chose language %s",
+		   content_type ? content_type : "(null)",
+		   filename,
+		   language ? gtk_source_language_get_id (language) : "(none)");
 
+	g_free (content_type);
+	g_free (text);
 	return language;
 }
 
@@ -470,7 +414,7 @@ get_language (GtkTextBuffer *buffer, const gchar *filename)
 	}
 
 	if (!language)
-		language = get_language_for_file (filename);
+		language = get_language_for_file (buffer, filename);
 
 	g_free (text);
 	return language;
@@ -574,6 +518,17 @@ margin_toggled_cb (GtkAction *action, gpointer user_data)
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+hl_bracket_toggled_cb (GtkAction *action, gpointer user_data)
+{
+	GtkTextBuffer *buffer;
+	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	buffer = gtk_text_view_get_buffer (user_data);
+	gtk_source_buffer_set_highlight_matching_brackets (
+		GTK_SOURCE_BUFFER (buffer),
+		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
+}
+
 static void
 hl_line_toggled_cb (GtkAction *action, gpointer user_data)
 {
@@ -583,6 +538,22 @@ hl_line_toggled_cb (GtkAction *action, gpointer user_data)
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+draw_spaces_toggled_cb (GtkAction *action, gpointer user_data)
+{
+	gboolean draw_spaces;
+
+	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	draw_spaces = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
+
+	if (draw_spaces)
+		gtk_source_view_set_draw_spaces (GTK_SOURCE_VIEW (user_data),
+						 GTK_SOURCE_DRAW_SPACES_ALL);
+	else
+		gtk_source_view_set_draw_spaces (GTK_SOURCE_VIEW (user_data),
+						 0);
+}
+
 static void
 wrap_lines_toggled_cb (GtkAction *action, gpointer user_data)
 {
@@ -662,24 +633,86 @@ new_view_cb (GtkAction *action, gpointer user_data)
 	gtk_widget_show (window);
 }
 
+static void
+forward_string_cb (GtkAction *action,
+		   gpointer   user_data)
+{
+	GtkSourceBuffer *buffer;
+	GtkSourceView *view;
+	GtkTextIter iter;
+	GtkTextMark *insert;
+
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
+
+	view = GTK_SOURCE_VIEW (user_data);
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
+	insert = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (buffer));
+
+	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (buffer),
+	                                  &iter,
+	                                  insert);
+
+	if (gtk_source_buffer_iter_forward_to_context_class_toggle (buffer,
+	                                                            &iter,
+	                                                            "string"))
+	{
+		gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (buffer), &iter);
+		gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (view), insert);
+	}
+}
+
+static void
+backward_string_cb (GtkAction *action,
+		    gpointer   user_data)
+{
+	GtkSourceBuffer *buffer;
+	GtkSourceView *view;
+	GtkTextIter iter;
+	GtkTextMark *insert;
+
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
+
+	view = GTK_SOURCE_VIEW (user_data);
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
+	insert = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (buffer));
+
+	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (buffer),
+	                                  &iter,
+	                                  insert);
+
+	if (gtk_source_buffer_iter_backward_to_context_class_toggle (buffer,
+	                                                             &iter,
+	                                                             "string"))
+	{
+		gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (buffer), &iter);
+		gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (view), insert);
+	}
+}
 
 /* Buffer action callbacks ------------------------------------------------------------ */
 
+static struct {
+	char *what;
+	char *replacement;
+	GtkSourceSearchFlags flags;
+} search_data = {
+	NULL,
+	NULL,
+	GTK_SOURCE_SEARCH_CASE_INSENSITIVE
+};
+
 static gboolean
-replace_dialog (GtkWidget *widget,
-		char     **what_p,
-		char     **replacement_p)
+search_dialog (GtkWidget            *widget,
+	       gboolean              replace,
+	       char                **what_p,
+	       char                **replacement_p,
+	       GtkSourceSearchFlags *flags_p)
 {
 	GtkWidget *dialog;
 	GtkEntry *entry1, *entry2;
-	static char *what, *replacement;
+	GtkToggleButton *case_sensitive;
 
-	if (!what)
-		what = g_strdup ("gtk");
-	if (!replacement)
-		replacement = g_strdup ("boo");
-
-	dialog = gtk_dialog_new_with_buttons ("Replace",
+	dialog = gtk_dialog_new_with_buttons (replace ? "Replace" : "Find",
 					      GTK_WINDOW (gtk_widget_get_toplevel (widget)),
 					      GTK_DIALOG_MODAL,
 					      GTK_STOCK_CANCEL,
@@ -691,18 +724,26 @@ replace_dialog (GtkWidget *widget,
 
 	entry1 = g_object_new (GTK_TYPE_ENTRY,
 			       "visible", TRUE,
-			       "text", what ? what : "",
+			       "text", search_data.what ? search_data.what : "",
 			       "activates-default", TRUE,
 			       NULL);
-	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
-				     GTK_WIDGET (entry1));
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
+				     GTK_WIDGET (entry1), TRUE, TRUE, 0);
 	entry2 = g_object_new (GTK_TYPE_ENTRY,
-			       "visible", TRUE,
-			       "text", replacement ? replacement : "",
+			       "visible", replace,
+			       "text", search_data.replacement ? search_data.replacement : "",
 			       "activates-default", TRUE,
 			       NULL);
-	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
-				     GTK_WIDGET (entry2));
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
+				     GTK_WIDGET (entry2), TRUE, TRUE, 0);
+
+	case_sensitive = g_object_new (GTK_TYPE_CHECK_BUTTON,
+				       "visible", TRUE,
+				       "label", "Case sensitive",
+				       "active", !(search_data.flags & GTK_SOURCE_SEARCH_CASE_INSENSITIVE),
+				       NULL);
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
+			    GTK_WIDGET (case_sensitive), FALSE, FALSE, 0);
 
 	while (TRUE)
 	{
@@ -716,47 +757,84 @@ replace_dialog (GtkWidget *widget,
 			break;
 	}
 
-	g_free (what);
-	*what_p = what = g_strdup (gtk_entry_get_text (entry1));
-	g_free (replacement);
-	*replacement_p = replacement = g_strdup (gtk_entry_get_text (entry2));
+	g_free (search_data.what);
+	*what_p = search_data.what = g_strdup (gtk_entry_get_text (entry1));
+	g_free (search_data.replacement);
+	*replacement_p = search_data.replacement = g_strdup (gtk_entry_get_text (entry2));
+	*flags_p = search_data.flags = gtk_toggle_button_get_active (case_sensitive) ?
+					0 : GTK_SOURCE_SEARCH_CASE_INSENSITIVE;
 
 	gtk_widget_destroy (dialog);
 	return TRUE;
 }
 
 static void
-debug_thing_3_cb (GtkAction *action,
-		  gpointer   user_data)
+do_search_replace (GtkTextView *view,
+		   gboolean     replace)
 {
-	GtkTextView *view = user_data;
 	GtkTextBuffer *buffer = gtk_text_view_get_buffer (view);
 	GtkTextIter iter;
 	char *what, *replacement;
+	GtkSourceSearchFlags flags;
 
-	if (!replace_dialog (GTK_WIDGET (view), &what, &replacement))
+	if (!search_dialog (GTK_WIDGET (view), replace, &what, &replacement, &flags))
 		return;
 
-	gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
+	if (replace)
+	{
+		gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
 
-	while (TRUE)
+		while (TRUE)
+		{
+			GtkTextIter match_start, match_end;
+
+			if (!gtk_source_iter_forward_search (&iter, what, flags,
+							     &match_start,
+							     &match_end,
+							     NULL))
+			{
+				break;
+			}
+
+			gtk_text_buffer_delete (buffer, &match_start, &match_end);
+			gtk_text_buffer_insert (buffer, &match_start, replacement, -1);
+			iter = match_start;
+		}
+	}
+	else
 	{
 		GtkTextIter match_start, match_end;
 
-		if (!gtk_text_iter_forward_search (&iter, what, 0,
-						   &match_start,
-						   &match_end,
-						   NULL))
-		{
-			break;
-		}
+		gtk_text_buffer_get_iter_at_mark (buffer, &iter, gtk_text_buffer_get_insert (buffer));
 
-		gtk_text_buffer_delete (buffer, &match_start, &match_end);
-		gtk_text_buffer_insert (buffer, &match_start, replacement, -1);
-		iter = match_start;
+		if (gtk_source_iter_forward_search (&iter, what, flags, &match_start, &match_end, NULL))
+		{
+			gtk_text_buffer_select_range (buffer, &match_start, &match_end);
+		}
+		else
+		{
+			GtkTextIter insert = iter;
+			gtk_text_buffer_get_start_iter (buffer, &iter);
+			if (gtk_source_iter_forward_search (&iter, what, flags, &match_start, &match_end, &insert))
+				gtk_text_buffer_select_range (buffer, &match_start, &match_end);
+		}
 	}
 }
 
+static void
+find_cb (GtkAction *action,
+	 gpointer   user_data)
+{
+	do_search_replace (user_data, FALSE);
+}
+
+static void
+replace_cb (GtkAction *action,
+	    gpointer   user_data)
+{
+	do_search_replace (user_data, TRUE);
+}
+
 static void
 open_file_cb (GtkAction *action, gpointer user_data)
 {
@@ -805,12 +883,12 @@ open_file_cb (GtkAction *action, gpointer user_data)
 #ifndef NON_BLOCKING_PAGINATION
 
 static void
-begin_print (GtkPrintOperation        *operation, 
+begin_print (GtkPrintOperation        *operation,
 	     GtkPrintContext          *context,
 	     GtkSourcePrintCompositor *compositor)
 {
 	gint n_pages;
-	
+
 	while (!gtk_source_print_compositor_paginate (compositor, context))
 		;
 
@@ -821,27 +899,27 @@ begin_print (GtkPrintOperation        *operation,
 #else
 
 static gboolean
-paginate (GtkPrintOperation        *operation, 
+paginate (GtkPrintOperation        *operation,
 	  GtkPrintContext          *context,
 	  GtkSourcePrintCompositor *compositor)
 {
 	g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
-	
+
 	if (gtk_source_print_compositor_paginate (compositor, context))
 	{
 		gint n_pages;
 
 		g_assert (gtk_source_print_compositor_get_pagination_progress (compositor) == 1.0);
 		g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
-		        
+
 		n_pages = gtk_source_print_compositor_get_n_pages (compositor);
 		gtk_print_operation_set_n_pages (operation, n_pages);
 
 
-		
+
 		return TRUE;
 	}
-     
+
 	return FALSE;
 }
 
@@ -865,11 +943,11 @@ draw_page (GtkPrintOperation        *operation,
 	PangoFontDescription *desc;
 	PangoRectangle rect;
 
-	
+
 	cr = gtk_print_context_get_cairo_context (context);
-	
+
 	cairo_save (cr);
-	
+
 	layout = gtk_print_context_create_pango_layout (context);
 
 	pango_layout_set_text (layout, "Draft", -1);
@@ -878,15 +956,15 @@ draw_page (GtkPrintOperation        *operation,
 	pango_layout_set_font_description (layout, desc);
 	pango_font_description_free (desc);
 
-	
+
 	pango_layout_get_extents (layout, NULL, &rect);
-		
-  	cairo_move_to (cr, 
+
+  	cairo_move_to (cr,
   		       (gtk_print_context_get_width (context) - ((double) rect.width / (double) PANGO_SCALE)) / 2,
   		       (gtk_print_context_get_height (context) - ((double) rect.height / (double) PANGO_SCALE)) / 2);
-	  		       
+
 	pango_cairo_layout_path (cr, layout);
-	
+
   	/* Font Outline */
 	cairo_set_source_rgba (cr, 0.85, 0.85, 0.85, 0.80);
 	cairo_set_line_width (cr, 0.5);
@@ -895,7 +973,7 @@ draw_page (GtkPrintOperation        *operation,
 	/* Font Fill */
 	cairo_set_source_rgba (cr, 0.8, 0.8, 0.8, 0.60);
 	cairo_fill (cr);
-	  	  
+
 	g_object_unref (layout);
 	cairo_restore (cr);
 #endif
@@ -905,7 +983,7 @@ draw_page (GtkPrintOperation        *operation,
 }
 
 static void
-end_print (GtkPrintOperation        *operation, 
+end_print (GtkPrintOperation        *operation,
 	   GtkPrintContext          *context,
 	   GtkSourcePrintCompositor *compositor)
 {
@@ -916,12 +994,12 @@ end_print (GtkPrintOperation        *operation,
 #define HEADER_FONT_NAME	"Sans 11"
 #define FOOTER_FONT_NAME	"Sans 11"
 #define BODY_FONT_NAME		"Monospace 9"
-	
+
 /*
-#define SETUP_FROM_VIEW 
+#define SETUP_FROM_VIEW
 */
 
-#undef SETUP_FROM_VIEW 
+#undef SETUP_FROM_VIEW
 
 
 static void
@@ -930,23 +1008,23 @@ print_file_cb (GtkAction *action, gpointer user_data)
 	GtkSourceView *view;
 	GtkSourceBuffer *buffer;
 	GtkSourcePrintCompositor *compositor;
-	GtkPrintOperation *operation;	
+	GtkPrintOperation *operation;
 	const gchar *filename;
 	gchar *basename;
-	
+
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
-	
+
 	view = GTK_SOURCE_VIEW (user_data);
 
-	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))); 
-	
-	filename = g_object_get_data (G_OBJECT (buffer), "filename");	
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
+
+	filename = g_object_get_data (G_OBJECT (buffer), "filename");
 	basename = g_filename_display_basename (filename);
-	
+
 #ifdef SETUP_FROM_VIEW
 	compositor = gtk_source_print_compositor_new_from_view (view);
 #else
-	
+
 	compositor = gtk_source_print_compositor_new (buffer);
 
 	gtk_source_print_compositor_set_tab_width (compositor,
@@ -984,26 +1062,26 @@ print_file_cb (GtkAction *action, gpointer user_data)
 
 	gtk_source_print_compositor_set_footer_font_name (compositor,
 							  FOOTER_FONT_NAME);
-#endif	
+#endif
 	operation = gtk_print_operation_new ();
-			
+
 	gtk_print_operation_set_job_name (operation, basename);
-	
+
 	gtk_print_operation_set_show_progress (operation, TRUE);
 
-#ifndef NON_BLOCKING_PAGINATION	
-  	g_signal_connect (G_OBJECT (operation), "begin-print", 
+#ifndef NON_BLOCKING_PAGINATION
+  	g_signal_connect (G_OBJECT (operation), "begin-print",
 			  G_CALLBACK (begin_print), compositor);
 #else
-  	g_signal_connect (G_OBJECT (operation), "paginate", 
+  	g_signal_connect (G_OBJECT (operation), "paginate",
 			  G_CALLBACK (paginate), compositor);
 #endif
-	g_signal_connect (G_OBJECT (operation), "draw-page", 
+	g_signal_connect (G_OBJECT (operation), "draw-page",
 			  G_CALLBACK (draw_page), compositor);
-	g_signal_connect (G_OBJECT (operation), "end-print", 
+	g_signal_connect (G_OBJECT (operation), "end-print",
 			  G_CALLBACK (end_print), compositor);
 
-	gtk_print_operation_run (operation, 
+	gtk_print_operation_run (operation,
 				 GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
 				 NULL, NULL);
 
@@ -1021,6 +1099,9 @@ update_cursor_position (GtkTextBuffer *buffer, gpointer user_data)
 	GtkTextIter iter, start;
 	GtkSourceView *view;
 	GtkLabel *pos_label;
+	gchar **classes;
+	gchar **classes_ptr;
+	GString *str;
 
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
 
@@ -1051,9 +1132,30 @@ update_cursor_position (GtkTextBuffer *buffer, gpointer user_data)
 		gtk_text_iter_forward_char (&start);
 	}
 
-	msg = g_strdup_printf ("char: %d, line: %d, column: %d", chars, row, col);
+	classes = gtk_source_buffer_get_context_classes_at_iter (GTK_SOURCE_BUFFER (buffer),
+	                                                         &iter);
+
+	str = g_string_new ("");
+	classes_ptr = classes;
+
+	while (classes_ptr && *classes_ptr)
+	{
+		if (classes_ptr != classes)
+		{
+			g_string_append (str, ", ");
+		}
+
+		g_string_append_printf (str, "%s", *classes_ptr);
+		++classes_ptr;
+	}
+
+	g_strfreev (classes);
+
+	msg = g_strdup_printf ("char: %d, line: %d, column: %d, classes: %s", chars, row, col, str->str);
 	gtk_label_set_text (pos_label, msg);
-      	g_free (msg);
+
+	g_free (msg);
+	g_string_free (str, TRUE);
 }
 
 static void
@@ -1103,80 +1205,100 @@ window_deleted_cb (GtkWidget *widget, GdkEvent *ev, gpointer user_data)
 	return TRUE;
 }
 
-static gboolean
-button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
+static void
+line_mark_activated (GtkSourceGutter *gutter, 
+                     GtkTextIter     *iter,
+                     GdkEventButton  *ev,
+                     GtkSourceView   *view)
 {
-	GtkSourceView *view;
 	GtkSourceBuffer *buffer;
+	GSList *mark_list;
+	const gchar *mark_type;
 
-	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (widget), FALSE);
+	if (ev->button == 1)
+		mark_type = MARK_TYPE_1;
+	else
+		mark_type = MARK_TYPE_2;
 
-	view = GTK_SOURCE_VIEW (widget);
 	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
 
-	if (!gtk_source_view_get_show_line_marks (view))
-		return FALSE;
+	/* get the marks already in the line */
+	mark_list = gtk_source_buffer_get_source_marks_at_line (buffer,
+								gtk_text_iter_get_line (iter),
+								mark_type);
 
-	/* check that the click was on the left gutter */
-	if (ev->window == gtk_text_view_get_window (GTK_TEXT_VIEW (view),
-						    GTK_TEXT_WINDOW_LEFT))
+	if (mark_list != NULL)
 	{
-		gint y_buf;
-		GtkTextIter line_start;
-		GSList *mark_list;
-		const gchar *mark_type;
-
-		if (ev->button == 1)
-			mark_type = MARK_TYPE_1;
-		else
-			mark_type = MARK_TYPE_2;
-
-		gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
-						       GTK_TEXT_WINDOW_LEFT,
-						       ev->x, ev->y,
-						       NULL, &y_buf);
-
-		/* get line bounds */
-		gtk_text_view_get_line_at_y (GTK_TEXT_VIEW (view),
-					     &line_start,
-					     y_buf,
-					     NULL);
-
-		/* get the marks already in the line */
-		mark_list = gtk_source_buffer_get_source_marks_at_line (buffer,
-									gtk_text_iter_get_line (&line_start),
-									mark_type);
-
-		if (mark_list != NULL)
-		{
-			/* just take the first and delete it */
-			gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (buffer),
-						     GTK_TEXT_MARK (mark_list->data));
-		}
-		else
-		{
-			/* no mark found: create one */
-			gtk_source_buffer_create_source_mark (buffer,
-							      NULL,
-							      mark_type,
-							      &line_start);
-		}
-
-		g_slist_free (mark_list);
+		/* just take the first and delete it */
+		gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (buffer),
+					     GTK_TEXT_MARK (mark_list->data));
+	}
+	else
+	{
+		/* no mark found: create one */
+		gtk_source_buffer_create_source_mark (buffer,
+						      NULL,
+						      mark_type,
+						      iter);
 	}
 
-	return FALSE;
+	g_slist_free (mark_list);
 }
 
 
 /* Window creation functions -------------------------------------------------------- */
 
+static gchar *
+mark_tooltip_func (GtkSourceMark *mark,
+		   gpointer	  user_data)
+{
+	GtkTextBuffer *buf;
+	GtkTextIter iter;
+	gint line, column;
+	
+	buf = gtk_text_mark_get_buffer (GTK_TEXT_MARK (mark));
+	
+	gtk_text_buffer_get_iter_at_mark (buf, &iter, GTK_TEXT_MARK (mark));
+	line = gtk_text_iter_get_line (&iter) + 1;
+	column = gtk_text_iter_get_line_offset (&iter);
+
+	if (strcmp (gtk_source_mark_get_category (mark), MARK_TYPE_1) == 0)
+		return g_strdup_printf ("Line: %d, Column: %d", line, column);
+	else
+		return g_strdup_printf ("<b>Line</b>: %d\n<i>Column</i>: %d", line, column);
+}
+
+static void
+add_source_mark_pixbufs (GtkSourceView *view)
+{
+	GdkColor color;
+
+	gdk_color_parse ("lightgreen", &color);
+	gtk_source_view_set_mark_category_background (view, MARK_TYPE_1, &color);
+	gtk_source_view_set_mark_category_icon_from_stock (view, MARK_TYPE_1, GTK_STOCK_YES);
+	gtk_source_view_set_mark_category_priority (view, MARK_TYPE_1, 1);
+	gtk_source_view_set_mark_category_tooltip_func (view,
+							MARK_TYPE_1,
+							mark_tooltip_func,
+							NULL,
+							NULL);
+
+	gdk_color_parse ("pink", &color);
+	gtk_source_view_set_mark_category_background (view, MARK_TYPE_2, &color);
+	gtk_source_view_set_mark_category_icon_from_stock (view, MARK_TYPE_2, GTK_STOCK_NO);
+	gtk_source_view_set_mark_category_priority (view, MARK_TYPE_2, 2);
+	gtk_source_view_set_mark_category_tooltip_markup_func (view,
+							       MARK_TYPE_2,
+							       mark_tooltip_func,
+							       NULL,
+							       NULL);
+}
+
 static GtkWidget *
 create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 {
 	GtkWidget *window, *sw, *view, *vbox, *pos_label, *menu;
 	PangoFontDescription *font_desc = NULL;
-	GdkPixbuf *pixbuf;
 	GtkAccelGroup *accel_group;
 	GtkActionGroup *action_group;
 	GtkUIManager *ui_manager;
@@ -1193,13 +1315,13 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 
 	/* view */
 	view = gtk_source_view_new_with_buffer (buffer);
-
+	
 	if (style_scheme)
 		gtk_source_buffer_set_style_scheme (buffer, style_scheme);
 
 	g_signal_connect (buffer, "mark-set", G_CALLBACK (move_cursor_cb), view);
 	g_signal_connect (buffer, "changed", G_CALLBACK (update_cursor_position), view);
-	g_signal_connect (view, "button-press-event", G_CALLBACK (button_press_cb), NULL);
+	g_signal_connect (view, "line-mark-activated", G_CALLBACK (line_mark_activated), view);
 	g_signal_connect (window, "delete-event", (GCallback) window_deleted_cb, view);
 
 	/* action group and UI manager */
@@ -1260,7 +1382,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		gtk_widget_modify_font (view, font_desc);
 		pango_font_description_free (font_desc);
 	}
-
+	
 	/* change view attributes to match those of from */
 	if (from)
 	{
@@ -1311,34 +1433,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		g_free (tmp);
 	}
 
-	/* add source mark pixbufs */
-	error = NULL;
-	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-green.png", &error)))
-	{
-		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_1, pixbuf);
-		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_1, 1);
-		g_object_unref (pixbuf);
-	}
-	else
-	{
-		g_message ("could not load source mark 1 image. Spurious messages might get triggered: %s\n",
-		error->message);
-		g_error_free (error);
-	}
-
-	error = NULL;
-	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-red.png", &error)))
-	{
-		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_2, pixbuf);
-		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_2, 2);
-		g_object_unref (pixbuf);
-	}
-	else
-	{
-		g_message ("could not load source mark 2 image. Spurious messages might get triggered: %s\n",
-		error->message);
-		g_error_free (error);
-	}
+	add_source_mark_pixbufs (GTK_SOURCE_VIEW (view));
 
 	gtk_widget_show_all (vbox);
 
@@ -1379,6 +1474,9 @@ create_main_window (GtkSourceBuffer *buffer)
 	/* retrieve the view action group at position 0 in the list */
 	action_group = g_list_nth_data (groups, 0);
 
+	action = gtk_action_group_get_action (action_group, "HlBracket");
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+
 	action = gtk_action_group_get_action (action_group, "ShowNumbers");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
@@ -1503,12 +1601,10 @@ main (int argc, char *argv[])
 
 	GOptionEntry entries[] = {
 	  { "style-scheme", 's', 0, G_OPTION_ARG_STRING, &style_scheme_id, "Style scheme name to use", "SCHEME"},
-	  { "default-paths", 'd', 0, G_OPTION_ARG_NONE, &use_default_paths, "Style scheme name to use", "SCHEME"},
+	  { "default-paths", 'd', 0, G_OPTION_ARG_NONE, &use_default_paths, "Use default search paths", NULL},
 	  { NULL }
 	};
 
-	g_thread_init (NULL);
-
 #ifdef TEST_XML_MEM
 	init_mem_stuff ();
 #endif
@@ -1520,10 +1616,6 @@ main (int argc, char *argv[])
 
 // 	gdk_window_set_debug_updates (TRUE);
 
-#ifdef USE_GNOME_VFS
-	gnome_vfs_init ();
-#endif
-
 	/* we do not use defaults so we don't need to install the library */
 	lang_dirs = use_default_paths ? NULL : builtin_lang_dirs;
 	lm = gtk_source_language_manager_get_default ();
@@ -1593,9 +1685,5 @@ main (int argc, char *argv[])
 
 	g_free (style_scheme_id);
 
-#ifdef USE_GNOME_VFS
-	gnome_vfs_shutdown ();
-#endif
-
 	return 0;
 }
